/*
**  Copyright (c) 2004 Sendmail, Inc. and its suppliers.
**	All rights reserved.
**  Sendmail, Inc. Confidential
**
**  $Id: util.c,v 1.7 2004/08/23 06:46:54 msk Exp $
*/

#ifndef lint
static char util_c_id[] = "@(#)$Id: util.c,v 1.7 2004/08/23 06:46:54 msk Exp $";
#endif /* !lint */

/* system includes */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

/* libsm includes */
#include <sm/string.h>

/* sid-filter includes */
#include "sid-filter.h"
#include "util.h"

/*
**  SID_SETMAXFD -- increase the file descriptor limit as much as possible
**
**  Parameters:
**  	None.
**
**  Return value:
**  	None.
*/

void
sid_setmaxfd()
{
	struct rlimit rlp;

	if (getrlimit(RLIMIT_NOFILE, &rlp) != 0)
	{
		syslog(LOG_WARNING, "getrlimit(): %s", strerror(errno));
	}
	else
	{
		rlp.rlim_cur = rlp.rlim_max;
		if (setrlimit(RLIMIT_NOFILE, &rlp) != 0)
		{
			syslog(LOG_WARNING, "setrlimit(): %s",
			       strerror(errno));
		}
	}
}

/*
**  SID_STRIPBRACKETS -- remove angle brackets from the sender address
**
**  Parameters:
** 	addr -- address to be processed
**
**  Return value:
**  	None.
*/

void
sid_stripbrackets(addr)
	char *addr;
{
	char *p, *q;

	assert(addr != NULL);

	p = addr;
	q = addr + strlen(addr) - 1;

	while (*p == '<' && *q == '>')
	{
		p++;
		*q-- = '\0';
	}

	if (p != addr)
	{
		for (q = addr; *p != '\0'; p++, q++)
			*q = *p;
		*q = '\0';
	}
}

/*
**  SID_LOWERCASE -- lowercase-ize a string
**
**  Parameters:
**  	str -- string to convert
**
**  Return value:
**  	None.
*/

void
sid_lowercase(str)
	char *str;
{
	char *p;

	assert(str != NULL);

	for (p = str; *p != '\0'; p++)
	{
		if (isascii(*p) && isupper(*p))
			*p = tolower(*p);
	}
}

/*
**  SID_LIST_LOOKUP -- look up a name in a peerlist
**
**  Parameters:
**  	list -- list of records to check
** 	data -- record to find
**
**  Return value:
**   	TRUE if found, FALSE otherwise
*/

static bool
sid_list_lookup(Peer list, char *data)
{
	Peer current;

	assert(list != NULL);
	assert(data != NULL);

	for (current = list; current != NULL; current = current->peer_next)
	{
		if (strcasecmp(data, current->peer_info) == 0)
			return TRUE;
	}

	return FALSE;
}

/*
**  SID_CHECKHOST -- check the peerlist for a host and its wildcards
**
**  Parameters:
**  	list -- list of records to check
**  	host -- hostname to find
**
**  Return value:
**  	None.
*/

bool
sid_checkhost(Peer list, char *host)
{
	char *p;

	assert(list != NULL);
	assert(host != NULL);

	if (sid_list_lookup(list, host))
		return TRUE;

	for (p = strchr(host, '.');
	     p != NULL;
	     p = strchr(p + 1, '.'))
	{
		if (sid_list_lookup(list, p))
			return TRUE;
	}

	return FALSE;
}

/*
**  SID_CHECKIP -- check a peerlist table for an IP address or its matching
**                 wildcards
**
**  Parameters:
**  	list -- list to check
**  	ip -- IP address to find
**
**  Return value:
**  	None.
*/

bool
sid_checkip(Peer list, struct sockaddr *ip)
{
	int c;
	char ipbuf[MAXHOSTNAMELEN + 1];
	struct in_addr addr;
	struct in_addr mask;

	assert(list != NULL);
	assert(ip != NULL);

	memcpy(&addr.s_addr, &((struct sockaddr_in *)ip)->sin_addr,
	       sizeof(addr.s_addr));

	/* test the IP by itself */
	memset(ipbuf, '\0', MAXHOSTNAMELEN + 1);
	snprintf(ipbuf, MAXHOSTNAMELEN, "%s", inet_ntoa(addr));
	if (sid_list_lookup(list, ipbuf))
		return TRUE;

	/* test the IP/32 */
	snprintf(ipbuf, MAXHOSTNAMELEN, "%s/32", inet_ntoa(addr));
	if (sid_list_lookup(list, ipbuf))
		return TRUE;

	/*
	**  Prepare to cycle through the other IP/bits possibilities.
	**  Turn on all the possible network bits in our bitmask.
	*/

	mask.s_addr = 0;
	for (c = 0; c < 32; c++)
		mask.s_addr |= htonl(1 << (31 - c));

	/* decompose the address bit-by-bit and try the appropriate entry */
	for (c = 0; c < 32; c++)
	{
		mask.s_addr ^= htonl(1 << c);
		addr.s_addr = (addr.s_addr & mask.s_addr);

		memset(ipbuf, '\0', MAXHOSTNAMELEN + 1);
		snprintf(ipbuf, MAXHOSTNAMELEN, "%s/%d", inet_ntoa(addr),
		         31 - c);

		if (sid_list_lookup(list, ipbuf))
			return TRUE;
	}

	return FALSE;
}
